/**
* @file MIDfile.cpp
* @brief MIDI/SMFt@Ĉq^v
* @author T.Modegi
* @date  2019.1.22
*/
#include "StdAfx.h"
#include <stdio.h>
#include "common.h"
#include "commparam.h"
#include "MIDfile.h"
#include "MIDdraw.h"
#include "MIDcnv_s1_fft.h"
#include "MIDcnv_s4_mid.h"
#include "commdata.h"

unsigned char *midbuf;
int pnt_midbuf=0,num_midbuf=0;
int stanza_unit=0;
FILE *midi_fp;
MIDIMSG *MidiMsg1;
DWORD song_len1=0;
MIDIMSG *MidiMsg2;
DWORD song_len2=0;

/**
* @fn get_time_res
* @brief SMF`MIDIf[^莞ԉ𑜓x̏擾
* @param ()  Ȃ
* @return ԉ𑜓x
* @sa 
* @detail
*/
int get_time_res() 
{
	int i,t;
	union {
		int i4;
		unsigned char c4[4];
	} cnv4;
	cnv4.i4=0;
	for (i=0;i<3;i++)  cnv4.c4[i]=midbuf[pnt_midbuf+2-i];
	pnt_midbuf +=3;
	t=384000000/cnv4.i4;
	return(t);
}

/**
* @fn get_time_res
* @brief SMF`MIDIf[^f^^C̏擾
* @param ()  Ȃ
* @return f^^C
* @sa 
* @detail
*/
int get_delta_time()
{
	int i,t;

	t=0;
	for (i=0;i<4;i++) {
	if (midbuf[pnt_midbuf]<128) {
		t +=(int)midbuf[pnt_midbuf];
	    pnt_midbuf++;
		return(t);
	}
	else {
		t +=(int)(midbuf[pnt_midbuf]-128);
		t= t*128;
	    pnt_midbuf++;
	}
	}
	return(t);
}

/**
* @fn Read_SMFdata
* @brief SMF`f[^midbufǂݍMIDIf[^\MidiMsg\z
* @param ()  Ȃ
* @return Ȃ
* @sa get_time_res,get_delta_time
* @detail
*/
int Read_SMFdata()
{
	int i;
	int dt;
	int ch;
	int cnt;

	pnt_midbuf=0;
	song_len=0;
	abs_time=0;
	max_channel=1;

	for (i=1;i<num_midbuf;i++) {
		if (midbuf[i]==0x00 && midbuf[i+1]==0xff && midbuf[i+2]==0x51) {
			pnt_midbuf=i+4;
			break;
		}
	}
	time_unit=get_time_res();

	while (pnt_midbuf<num_midbuf) {
		dt=get_delta_time();
		if ((midbuf[pnt_midbuf]>=0x80 && midbuf[pnt_midbuf]<=0x9f) || 
				(midbuf[pnt_midbuf]>=0xe0 && midbuf[pnt_midbuf]<=0xef) ||
				(midbuf[pnt_midbuf]>=0xb0 && midbuf[pnt_midbuf]<=0xbf && 
				midbuf[pnt_midbuf+1]==0x0b)) {
				ch=midbuf[pnt_midbuf] & 0x0f;
				if (ch>max_channel-1) max_channel=ch+1;
				MidiMsg[song_len].com=midbuf[pnt_midbuf];
				MidiMsg[song_len].note=midbuf[pnt_midbuf+1];
				MidiMsg[song_len].vel=midbuf[pnt_midbuf+2];
				abs_time +=(DWORD)dt;
				MidiMsg[song_len].ms=abs_time;
				song_len++;
				pnt_midbuf +=3;
		}
		else if (midbuf[pnt_midbuf]>=0xc0 && midbuf[pnt_midbuf]<=0xcf) {
			pnt_midbuf +=2;
		}
		else if (midbuf[pnt_midbuf]>=0xb0 && midbuf[pnt_midbuf]<=0xbf) {
			if (midbuf[pnt_midbuf+1]==0 && midbuf[pnt_midbuf+4]==0x20) pnt_midbuf +=6;
			else pnt_midbuf +=3;
		}
		else if (midbuf[pnt_midbuf]==0xff) {
			cnt=(int)midbuf[pnt_midbuf+2];
			pnt_midbuf +=3+cnt;
		}
		else if (midbuf[pnt_midbuf]==0xf0) {
			cnt=(int)midbuf[pnt_midbuf+1];
			pnt_midbuf +=2+cnt;
		}
		else {
			break;
		}
	}

	total_time=abs_time*1000/time_unit;
	return(song_len);
}

/**
* @fn Read_SMFdata2
* @brief SMF`f[^midbufǂݍMIDIf[^\MidiMsg\ziwb_{̂̂݁j
* @param ()  Ȃ
* @return Ȃ
* @sa get_delta_time
* @detail
*/
int Read_SMFdata2()
{
	int dt;
	int ch;
	int cnt;

	pnt_midbuf=16;
	song_len=0;
	abs_time=0;

	while (pnt_midbuf<num_midbuf-4) {
		dt=get_delta_time();
		if ((midbuf[pnt_midbuf]>=0x80 && midbuf[pnt_midbuf]<=0x9f) || 
				(midbuf[pnt_midbuf]>=0xe0 && midbuf[pnt_midbuf]<=0xef) ||
				(midbuf[pnt_midbuf]>=0xb0 && midbuf[pnt_midbuf]<=0xbf && 
				midbuf[pnt_midbuf+1]==0x0b)) {
				ch=midbuf[pnt_midbuf] & 0x0f;
				if (ch>max_channel-1) max_channel=ch+1;
				MidiMsg[song_len].com=midbuf[pnt_midbuf];
				MidiMsg[song_len].note=midbuf[pnt_midbuf+1];
				MidiMsg[song_len].vel=midbuf[pnt_midbuf+2];
				abs_time +=(DWORD)dt;
				MidiMsg[song_len].ms=abs_time;
				song_len++;
				pnt_midbuf +=3;
		}
		else if (midbuf[pnt_midbuf]>=0xc0 && midbuf[pnt_midbuf]<=0xcf) {
			pnt_midbuf +=2;
		}
		else if (midbuf[pnt_midbuf]>=0xb0 && midbuf[pnt_midbuf]<=0xbf) {
			if (midbuf[pnt_midbuf+1]==0 && midbuf[pnt_midbuf+4]==0x20) pnt_midbuf +=6;
			else pnt_midbuf +=3;
		}
		else if (midbuf[pnt_midbuf]==0xff) {
			cnt=(int)midbuf[pnt_midbuf+2];
			pnt_midbuf +=3+cnt;
		}
		else if (midbuf[pnt_midbuf]==0xf0) {
			cnt=(int)midbuf[pnt_midbuf+1];
			pnt_midbuf +=2+cnt;
		}
		else {
			break;
		}
	}

	total_time=abs_time*1000/time_unit;
	return(song_len);
}

/**
* @fn Read_SMFfile
* @brief SMF`t@C̓ǂݍ
* @param (fname)  SMF`t@C
* @return őa
* @sa Read_SMFdata2,Read_SMFdata,count_harmony
* @detail Foramat1T|[gĂ邪AForamat1͌gpȂ
*/
int Read_SMFfile(char fname[])
{
	char buff[80];
	DWORD i,i1,i2;
	DWORD dwRead;
	HANDLE  hFile;
	int cur_harmony;
	union {
		int i4;
		unsigned char c4[4];
	} cnv4;

	hFile=CreateFile(fname,GENERIC_READ,0,NULL,OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if (hFile==INVALID_HANDLE_VALUE) return(-1);
	sprintf(fname_bck,"%s",fname);
	ReadFile(hFile,buff,22,&dwRead,NULL);

	if (song_len>0) free(MidiMsg);

	if (buff[9]==0x01) {

// format1 //
		ReadFile(hFile,buff,19,&dwRead,NULL);
		cnv4.i4=0;
		for (i=0;i<3;i++)  cnv4.c4[i]=buff[14-i];
		time_unit=384000000/cnv4.i4;
		max_channel=2;
// track 1
		ReadFile(hFile,buff,8,&dwRead,NULL);
		for (i=0;i<4;i++) cnv4.c4[i]=buff[7-i];
		num_midbuf=cnv4.i4;
		if (num_midbuf<=0) return(-1);
		midbuf=(unsigned char *)malloc(num_midbuf);
		ReadFile(hFile,midbuf,num_midbuf,&dwRead,NULL);
		MidiMsg=(MIDIMSG *)malloc(num_midbuf*sizeof(MIDIMSG)/4);
		song_len1=Read_SMFdata2();
		free(midbuf);
		MidiMsg1=(MIDIMSG *)malloc(song_len1*sizeof(MIDIMSG));
		for (i=0;i<song_len1;i++) {
			MidiMsg1[i].ms=MidiMsg[i].ms;
			MidiMsg1[i].com=MidiMsg[i].com; MidiMsg1[i].note=MidiMsg[i].note; 
			MidiMsg1[i].vel=MidiMsg[i].vel; MidiMsg1[i].chan=MidiMsg[i].chan;
		}
		free(MidiMsg);
// track 2
		ReadFile(hFile,buff,8,&dwRead,NULL);
		for (i=0;i<4;i++) cnv4.c4[i]=buff[7-i];
		num_midbuf=cnv4.i4;
		if (num_midbuf<=0) return(-1);
		midbuf=(unsigned char *)malloc(num_midbuf);
		ReadFile(hFile,midbuf,num_midbuf,&dwRead,NULL);
		MidiMsg=(MIDIMSG *)malloc(num_midbuf*sizeof(MIDIMSG)/4);
		song_len2=Read_SMFdata2();
		free(midbuf);
		MidiMsg2=(MIDIMSG *)malloc(song_len2*sizeof(MIDIMSG));
		for (i=0;i<song_len2;i++) {
			MidiMsg2[i].ms=MidiMsg[i].ms;
			MidiMsg2[i].com=MidiMsg[i].com; MidiMsg2[i].note=MidiMsg[i].note; 
			MidiMsg2[i].vel=MidiMsg[i].vel; MidiMsg2[i].chan=MidiMsg[i].chan;
		}
		free(MidiMsg);
// Gousei
		song_len=song_len1+song_len2;
		MidiMsg=(MIDIMSG *)malloc(song_len*sizeof(MIDIMSG));
		i=i1=i2=0;
		while (i<song_len) {
			if (i1<song_len1 && i2<song_len2) {
				if (MidiMsg1[i1].ms<=MidiMsg2[i2].ms) {
					MidiMsg[i].ms=MidiMsg1[i1].ms;
					MidiMsg[i].com=MidiMsg1[i1].com; MidiMsg[i].note=MidiMsg1[i1].note; 
					MidiMsg[i].vel=MidiMsg1[i1].vel; MidiMsg[i].chan=MidiMsg1[i1].chan;
					i++; i1++;
				}
				else if (MidiMsg1[i1].ms>MidiMsg2[i2].ms) {
					MidiMsg[i].ms=MidiMsg2[i2].ms;
					MidiMsg[i].com=MidiMsg2[i2].com; MidiMsg[i].note=MidiMsg2[i2].note; 
					MidiMsg[i].vel=MidiMsg2[i2].vel; MidiMsg[i].chan=MidiMsg2[i2].chan;
					i++; i2++;
				}
			}
			else if (i1<song_len1) {
					MidiMsg[i].ms=MidiMsg1[i1].ms;
					MidiMsg[i].com=MidiMsg1[i1].com; MidiMsg[i].note=MidiMsg1[i1].note; 
					MidiMsg[i].vel=MidiMsg1[i1].vel; MidiMsg[i].chan=MidiMsg1[i1].chan;
					i++; i1++;
			}
			else if (i2<song_len2) {
					MidiMsg[i].ms=MidiMsg2[i2].ms;
					MidiMsg[i].com=MidiMsg2[i2].com; MidiMsg[i].note=MidiMsg2[i2].note; 
					MidiMsg[i].vel=MidiMsg2[i2].vel; MidiMsg[i].chan=MidiMsg2[i2].chan;
					i++; i2++;
			}
		}
		free(MidiMsg1);
		free(MidiMsg2);

	}

// format 0 //
	else {
		for (i=0;i<4;i++) cnv4.c4[i]=buff[21-i];
		num_midbuf=cnv4.i4;
		if (num_midbuf<=0) return(-1);

		midbuf=(unsigned char *)malloc(num_midbuf);
		ReadFile(hFile,midbuf,num_midbuf,&dwRead,NULL);	
		MidiMsg=(MIDIMSG *)malloc(num_midbuf*sizeof(MIDIMSG)/4);
		song_len=Read_SMFdata();
		free(midbuf);
		max_channel=1;
	}
//
	CloseHandle(hFile);
	cur_harmony=count_harmony();

	return(cur_harmony);
}

/**
* @fn conv_time_res
* @brief ԉ𑜓xSMF`f[^ɕ
* @param (t)  ԉ𑜓x
* @param (c)  ꂽԉ𑜓xiRoCgj
* @return őa
* @sa 
* @detail 
*/
int conv_time_res(int t,unsigned char c[]) 
{
	int i;
	union {
		int i4;
		unsigned char c4[4];
	} cnv4;
	cnv4.i4=384000000/t;
	for (i=0;i<3;i++)  c[i]=cnv4.c4[i];
	return(cnv4.i4);
}

/**
* @fn write_MIDIbuffer
* @brief SMF`f[^zmidbufɏ
* @param (buff)  SMF`f[^
* @param (nbyte)  oCg
* @return 
* @sa 
* @detail 
*/
void write_MIDIbuffer(unsigned char buff[],int nbyte)
{
	int i;
	for (i=0;i<nbyte;i++) 
		midbuf[pnt_midbuf+i]=buff[i];
	pnt_midbuf += nbyte;
	if (pnt_midbuf>=num_midbuf) {
		MessageBox(NULL,"MIDIo̓obt@TCY𒴂܂B",
			"MIDIo̓t@CG[",MB_OK);
		pnt_midbuf=0;
	}
}

/**
* @fn write_MIDIbuffer
* @brief PoCgSMF`f[^zmidbufɏ
* @param (nbyte)  SMF`f[^
* @return 
* @sa 
* @detail 
*/
void write_MIDIbuffer(int nbyte)
{

	midbuf[pnt_midbuf]=(unsigned char)nbyte;
	pnt_midbuf ++;
	if (pnt_midbuf>=num_midbuf) {
		MessageBox(NULL,"MIDIo̓obt@TCY𒴂܂B",
			"MIDIo̓t@CG[",MB_OK);
		pnt_midbuf=0;
	}
}

/**
* @fn write_delta_time
* @brief f^^CSMF`f[^ɕϊĔzmidbufɏ
* @param (dtime)  f^^C
* @return 
* @sa write_MIDIbuffer
* @detail 
*/
void write_delta_time(int dtime)
{
	int i1,i2,i3;
	if (dtime<=0)  {
		write_MIDIbuffer(0);
	}
	else if (dtime<128) {
		write_MIDIbuffer(dtime);
	}
	else if(dtime<16384) {
		i1=dtime/128;  i2=dtime-i1*128;
		write_MIDIbuffer(i1+128);
		write_MIDIbuffer(i2);
	}
	else {
		i1=dtime/16384;  i2=(dtime-i1*16384)/128;
		i3=dtime-i1*16384-i2*128;
		write_MIDIbuffer(i1+128);
		write_MIDIbuffer(i2+128);
		write_MIDIbuffer(i3);
	}
}

/**
* @fn write_voice
* @brief {CXSMF`f[^ɕϊĔzmidbufɏ
* @param (chan) `l
* @param (bank1) PoNԍ
* @param (bank2) QoNԍ
* @param (prog) vOԍ
* @param (vol) ʏ
* @return 
* @sa write_delta_time,write_MIDIbuffer
* @detail 
*/
BOOL write_voice(int chan,int bank1,int bank2,int prog,int vol)
{
	unsigned char buff[3];
	if (prog<=0)  return(FALSE);
	buff[0]=0xb0+chan;
	buff[1]=0;
	buff[2]=bank1;
	write_delta_time(0);
	write_MIDIbuffer(buff,3);
	buff[0]=0x20;
	buff[1]=bank2;
	write_delta_time(0);
	write_MIDIbuffer(buff,2);
	buff[0]=0xc0+chan;
	buff[1]=prog-1;
	write_delta_time(0);
	write_MIDIbuffer(buff,2);
	buff[0]=0xb0+chan;
	buff[1]=0x07;
	buff[2]=vol;
	write_delta_time(0);
	write_MIDIbuffer(buff,3);
	return(TRUE);
}

/**
* @fn write_pitchbend
* @brief sb`xhEGNXvbṼZbgwSMF`f[^ɕϊĔzmidbufɏ
* @param (chan) `l
* @return 
* @sa write_delta_time,write_MIDIbuffer
* @detail 
*/
BOOL write_pitchbend(int chan)
{
	unsigned char buff[3];

// pitch bend reset
	buff[0]=0xe0+chan;
	buff[1]=0;
	buff[2]=0x40;
	write_delta_time(0);
	write_MIDIbuffer(buff,3);

// Add Expression Reset
	buff[0]=0xb0+chan;
	buff[1]=0x0b;
	buff[2]=0x7f;
	write_delta_time(0);
	write_MIDIbuffer(buff,3);
	return(TRUE);
}

/**
* @fn Write_SMFdata
* @brief MIDIf[^\MidiMsgSMF`f[^ɕϊĔzmidbufɏ
* @param (chan) `l
* @param (mode) ŏ̃`lގ0w肵At̏ʒu炷
* @return 
* @sa write_voice,write_pitchbend,write_delta_time
* @detail 
*/
int Write_SMFdata(int chan,int mode)
{
	DWORD i;
	int j;
	int dt;
	DWORD abst1,abst;
	int ch;

	pnt_midbuf=0;

	if (chan==0) {
		write_voice(0,MIDI_bank1,MIDI_bank2,MIDI_program,MIDI_volume);
		write_pitchbend(0);
	}
	else {
		if (max_channel==1) {
			write_voice(0,MIDI_bank1,MIDI_bank2,MIDI_program,MIDI_volume);
			write_pitchbend(0);
		}
		else {
			for (j=0;j<max_channel;j++) 
				write_voice(j,MIDI_bank1,MIDI_bank2,MIDI_program,MIDI_volume);
			for (j=0;j<max_channel;j++) write_pitchbend(j);
		}
	}

	abst1=0;
	for (i=0;i<song_len;i++) {
		ch=MidiMsg[i].com & 0x0f;
		if (chan<0 || ch==chan) {
		abst=MidiMsg[i].ms;
		dt=(int)(abst-abst1);
		if (dt<0) dt=0;
		write_delta_time(dt);
		midbuf[pnt_midbuf]=MidiMsg[i].com;
		midbuf[pnt_midbuf+1]=MidiMsg[i].note;
		midbuf[pnt_midbuf+2]=MidiMsg[i].vel;
		pnt_midbuf +=3;
		abst1 +=(DWORD)dt;
		}
	}
	abs_time=abst1;
	total_time=abs_time*1000/time_unit;
	if (mode==0) {
	abst=(abs_time+3071)/3072;
	dt=(int)(abst*3072-abs_time);
	write_delta_time(dt);
	}
	return(pnt_midbuf);
}

/**
* @fn Write_MIDfile
* @brief SMF`f[^midbufSMF`t@Cɏ
* @param (fname) SMF`t@C
* @return SMF`f[^̃oCg
* @sa Write_SMFdata
* @detail 
*/
/*  Write Standard MIDI File */
int Write_MIDfile(char fname[])
{
	int i;
	static unsigned char header1[14]={ 
		0x4d,0x54,0x68,0x64, 0x00,0x00,0x00,0x06,
		0x00,0x00, 0x00,0x01, 0x01,0x80 };
	static unsigned char track_start[4]={ 0x4d,0x54,0x72,0x6b};
	static unsigned char track_name1[10]={
		0x00, 0xff,0x03,0x06, 0xc4,0xd7,0xaf,0xb8,0x20,0x31 };
	static unsigned char setup_data1[15]={
		0x00, 0xff,0x58,0x04, 0x04,0x02,0x18,0x08,
		0x00, 0xff,0x51,0x03, 0x03,0xd0,0x90
	};
	static unsigned char track_end[3]={ 0xff,0x2f,0x00 };
	unsigned char length[4],ts[4];
	DWORD dwWrite;
	HANDLE  hFile;
	union {
		int i4;
		unsigned char c4[4];
	} cnv4;

	if (song_len<=0) return(-1);
	hFile=CreateFile(fname,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,
		FILE_ATTRIBUTE_NORMAL,NULL);
	if (hFile==INVALID_HANDLE_VALUE) return(-1);

	num_midbuf=song_len*12;
	midbuf=(unsigned char *)malloc(song_len*12);
 
/* Set Time Scale Parameter */
	conv_time_res(time_unit,ts);
	setup_data1[14]=ts[0];
	setup_data1[13]=ts[1];
	setup_data1[12]=ts[2];

/*  Write File */
	WriteFile(hFile,header1,14,&dwWrite,NULL);
	num_midbuf=Write_SMFdata(-1,0);
	cnv4.i4=num_midbuf+25+3;
	for (i=0;i<4;i++) length[i]=cnv4.c4[3-i];
	WriteFile(hFile,track_start,4,&dwWrite,NULL);
	WriteFile(hFile,length,4,&dwWrite,NULL);
	WriteFile(hFile,track_name1,10,&dwWrite,NULL);
	WriteFile(hFile,setup_data1,15,&dwWrite,NULL);
	WriteFile(hFile,midbuf,num_midbuf,&dwWrite,NULL);
	WriteFile(hFile,track_end,3,&dwWrite,NULL);

	CloseHandle(hFile);
	free(midbuf);
//	sprintf(fname_bck,"%s",fname);

	return(num_midbuf);
}


/**
* @fn conv_note_char
* @brief m[gio[𕶎ɕϊiRȉj
* @param (note) m[gio[
* @param (buf) ϊꂽ
* @return Ȃ
* @sa 
* @detail 
*/
void conv_note_char(int note,char buf[]) 
{
	int n1,n2;
	char code[3];
	n1=note/12;
	n2=note-n1*12;
	code[1]=code[2]='\0';
	if (n2==0 || n2==1)  code[0]='C';
	else if (n2==2 || n2==3)  code[0]='D';
	else if (n2==4)  code[0]='E';
	else if (n2==5 || n2==6)  code[0]='F';
	else if (n2==7 || n2==8)  code[0]='G';
	else if (n2==9 || n2==10)  code[0]='A';
	else if (n2==11)  code[0]='B';
	if (n2==1 || n2==3 || n2==6 || n2==8 || n2==10)  code[1]='#';
	sprintf(buf,"%s%d",code,n1-2);
}

/**
* @fn write_note_string_0
* @brief PMIDIm[gCxgt@Cɏ
* @param (note) m[gio[
* @param (dt) f^^C
* @param (vel) xVeB
* @return Ȃ
* @sa 
* @detail tH[}bgŌݖgp 
*/
void write_note_string_0(int note,int dt,int vel) 
{
	int n1,n2;
	char code[3];
	n1=note/12;
	n2=note-n1*12;
	code[1]=code[2]='\0';
	if (n2==0 || n2==1)  code[0]='C';
	else if (n2==2 || n2==3)  code[0]='D';
	else if (n2==4)  code[0]='E';
	else if (n2==5 || n2==6)  code[0]='F';
	else if (n2==7 || n2==8)  code[0]='G';
	else if (n2==9 || n2==10)  code[0]='A';
	else if (n2==11)  code[0]='B';
	if (n2==1 || n2==3 || n2==6 || n2==8 || n2==10)  code[1]='#';
	fprintf(midi_fp,"%s%d %d %d\n",code,n1-2,dt,vel);
}

/**
* @fn write_note_string_2
* @brief PMIDIm[gCxgt@Cɏ
* @param (note) m[gio[
* @param (vel) xVeB
* @return Ȃ
* @sa 
* @detail
*/
void write_note_string_2(int note,int vel) 
{
	int n1,n2;
	char code[3];
	n1=note/12;
	n2=note-n1*12;
	code[1]=code[2]='\0';
	if (n2==0 || n2==1)  code[0]='C';
	else if (n2==2 || n2==3)  code[0]='D';
	else if (n2==4)  code[0]='E';
	else if (n2==5 || n2==6)  code[0]='F';
	else if (n2==7 || n2==8)  code[0]='G';
	else if (n2==9 || n2==10)  code[0]='A';
	else if (n2==11)  code[0]='B';
	if (n2==1 || n2==3 || n2==6 || n2==8 || n2==10)  code[1]='#';
	fprintf(midi_fp,"%s%d %d\n",code,n1-2,vel);
}

/**
* @fn Write_TXTfile
* @brief MIDIf[^\MidiMsgTXT`MIDIt@Cɏ
* @param (fname) TXT`MIDIt@C
* @return MIDICxg̑
* @sa write_note_string_2
* @detail 
*/
int Write_TXTfile(char fname[])
{
	DWORD i;
	int dt;
	int vel;
	int com,note;
	int dat;
	int mc;
	DWORD abst,abst1;

	if (song_len<=0) return(0);
	midi_fp=fopen(fname,"w");
	if (midi_fp==NULL) return(0);

	fprintf(midi_fp,"Event_Max_count %d\n",song_len);
	fprintf(midi_fp,"Time_Resolution %d\n",time_unit);
	if (abs_timecode) mc= -max_channel;
	abst1=0;

	for (i=0;i<song_len;i++) {
		abst=MidiMsg[i].ms;
		if (abs_timecode) dt=(int)abst;
		else {
			dt=(int)(abst-abst1);
			if (dt<0) dt=0;
			abst1 +=(DWORD)dt;
		}
		vel=MidiMsg[i].vel;
		note=MidiMsg[i].note;
		com=MidiMsg[i].com;
		if (dt>0) fprintf(midi_fp,"T %d\n",dt);
		if (com>=0x80 && com<=0x9f) {
			if (com<0x90)  vel=0;
			write_note_string_2(note,vel);
		}
		else if (com>=0xe0 && com<=0xef) {
			dat=vel*128+note;
			fprintf(midi_fp,"PB%2x %d\n",com,dat);
		}
		else if (com>=0xb0 && com<=0xbf) {
			fprintf(midi_fp,"XP%2x %d\n",com,vel);
		}
	}
	abs_time=abst1;
	total_time=abs_time*1000/time_unit;

	sprintf(fname_bck,"%s",fname);
	fclose(midi_fp);
	return(song_len);
}

/**
* @fn conv_char_note
* @brief m[gio[ɕϊ
* @param (buf) 
* @param (com) CxgʃR}h 
* @return m[gio[
* @sa 
* @detail 
*/
int conv_char_note(char buf[],int *com) 
{
	int p,nn,oct,pol,tn;
	char num[3];
	*com=0;
	p=0;
	if (buf[p]=='C' || buf[p]=='c')  nn=0;
	else if (buf[p]=='D' || buf[p]=='d') nn=2;
	else if (buf[p]=='E' || buf[p]=='e') nn=4;
	else if (buf[p]=='F' || buf[p]=='f') nn=5;
	else if (buf[p]=='G' || buf[p]=='g') nn=7;
	else if (buf[p]=='A' || buf[p]=='a') nn=9;
	else if (buf[p]=='B' || buf[p]=='b') nn=11;
	else if (buf[p]=='P' || buf[p]=='p') {
		num[0]=buf[p+2];
		num[1]=buf[p+3];
		num[2]='\0';
		sscanf(num,"%2x",com);
		return(128);
	}
	else if (buf[p]=='X' || buf[p]=='x') {
		num[0]=buf[p+2];
		num[1]=buf[p+3];
		num[2]='\0';
		sscanf(num,"%2x",com);
		return(129);
	}
	else return(-1);
	p++;

	if (buf[p]=='#')  { nn++; p++; }
	else if (buf[p]=='b') { nn--; p++; }
	pol=1;
	if  (buf[p]=='-')  { pol= -1; p++; }

	num[0]=buf[p];
	num[1]=num[2]='\0';
	sscanf(num,"%d",&oct);
	tn=24+oct*12*pol+nn;
	return (tn);
}


/**
* @fn Read_MIDIraw
* @brief TXT`MIDIt@Cǂݍ݂MIDIf[^\MidiMsg\z
* @param (fname) TXT`MIDIt@C
* @return őa
* @sa conv_char_note,count_harmony
* @detail
*/
int Read_MIDIraw(char fname[])
{
	int i;
	FILE *fp;
	char buff[80],note[20];
	int vel,nn,com,dat;
	int sw;
	DWORD dt;
	DWORD size_buff;
	int cur_harmony;

	abs_time=0;
	fp=fopen(fname,"r");
	if (fp==NULL) return(0);
	sprintf(fname_bck,"%s",fname);

	fscanf(fp,"%s %d",buff,&size_buff);
	if (song_len>0) free(MidiMsg);
	song_len=0;
	MidiMsg=(MIDIMSG *)malloc(size_buff*sizeof(MIDIMSG));
	fscanf(fp,"%s %d",buff,&time_unit);
	for (i=0;i<20;i++) note[i]='\0';
	total_stanza=0;
	while(fscanf(fp,"%s %d",note,&vel)!=EOF) { 
		if (total_stanza<80 && note[0]=='/') {
			stanza[total_stanza]=song_len;
			total_stanza++;
		}
		else if (note[0]=='T' || note[0]=='t') {
			dt=vel;
			abs_time +=dt;
		}
		else {
			nn=conv_char_note(note,&sw);
			for (i=0;i<20;i++) note[i]='\0';
			if (nn!= -1) {
				if (nn==129) {
					MidiMsg[song_len].com=sw; MidiMsg[song_len].note=0x0b;
					MidiMsg[song_len].vel=vel; MidiMsg[song_len].chan=sw & 0x0f;
				}
				else if (nn==128) {
					dat=vel; vel=dat/128; nn=dat-vel*128;
					MidiMsg[song_len].com=sw; MidiMsg[song_len].note=nn;
					MidiMsg[song_len].vel=vel; MidiMsg[song_len].chan=sw & 0x0f;
				}
				else {
					if (vel>0) com=0x90; else com=0x80;
					MidiMsg[song_len].com=com; MidiMsg[song_len].note=nn;
					MidiMsg[song_len].vel=vel; MidiMsg[song_len].chan=0;
				}
				if (MidiMsg[song_len].com!=0) {
					MidiMsg[song_len].ms=abs_time;
					song_len++;
					if (song_len>=size_buff)  break;
				}
			}
		}
	}
	max_channel=1;
	fclose(fp);

	if (total_stanza<80)
		stanza[total_stanza]=song_len;
	total_time=abs_time*1000/time_unit;
	cur_harmony=count_harmony();

	return(cur_harmony);
}

/**
* @fn CloseMIDIFile
* @brief MIDIf[^\MidiMsgWAVEf[^zwavbufj
* @param ()  Ȃ
* @return Ȃ
* @sa 
* @detail
*/
void CloseMIDIFile()
{
	if (song_len>0) free(MidiMsg);
	song_len=0;
	if (num_sample>0)  free(wavbuf);
	num_sample=0;
}

/**
* @fn Change_FileExt
* @brief t@C̊gqύX
* @param (fname) ̓t@C
* @param (fname2) o̓t@C
* @param (ext) gq
* @return Ȃ
* @sa 
* @detail
*/
void Change_FileExt(char fname[],char fname2[],char ext[])
{
	int i,p,n;
	n=254;
	for (i=0;i<255;i++) {
		fname2[i]=fname[i];
		if (fname[i]=='\0') { n=i; break; }
	}
	p= -1;
	for (i=n;i>=0;i--) if (fname2[i]=='.') {p=i; break;}
	if (p>=0) {
		for (i=0;i<3;i++) fname2[p+i+1]=ext[i];
	}
}

/**
* @fn Change_FileExt
* @brief t@C"_M"ǉgqύX
* @param (fname) ̓t@C
* @param (fname2) o̓t@C
* @param (ext) gq
* @return Ȃ
* @sa 
* @detail
*/
void Change_FileExtM(char fname[],char fname2[],char ext[])
{
	int i,p,n;
	n=254;
	for (i=0;i<255;i++) {
		fname2[i]=fname[i];
		if (fname[i]=='\0') { n=i; break; }
	}
	p= -1;
	for (i=n;i>=0;i--) if (fname2[i]=='.') {p=i; break;}
	if (p>=0) {
		fname2[p+2]=fname2[p];
		fname2[p]='_';
		fname2[p+1]='M';
		p +=2;
		for (i=0;i<3;i++) fname2[p+i+1]=ext[i];
	}
	fname2[p+4]='\0';
}

