/**
* @file MIDout.cpp
* @brief  MIDIf[^̉t
* @author T.Modegi
* @date  2020.9.22
*/
#include "StdAfx.h"
#include <MMSYSTEM.h>
#include <stdio.h>
#include "common.h"
#include "commparam.h"
#include "commdata.h"
#include "MIDdraw.h"

enum Status
{
	StatusOn,
	StatusError,
	StatusOff,
	StatusSendSysExcl,
} eStatus=StatusOff;

MMRESULT rc;
UINT  nDevId=1;
UINT num_nDevId=0;
HMIDIOUT  hmo=NULL;
MIDIHDR pmh;

typedef struct
{
	DWORD dwNextMsg; // Next MIDI Short Message
	DWORD ms;  // current time position [msec]
} MIDISONG, *LPMIDISONG;

MIDIMSG *LPMIDIMSG;

MIDISONG MidiSong;
TIMECAPS tc;
UINT nTimerId;
UINT nTimerRes;

HWND hWnd_main2;
HDC hdc_main;

/**
* @fn Get_MIDI_device_list
* @brief gp\MIDIfoCX擾ăXgAbv
* @param (hDlg)  _CAOWindownh
* @param (hlist)  _CAÕXg{bNX̃nh
* @return Ȃ
* @sa 
* @detail@_CAÕXg{bNXɏo͂
*/
void Get_MIDI_device_list(HWND hDlg,UINT hlist) 
{
	UINT i;
	MIDIOUTCAPS moc;
	num_nDevId=midiOutGetNumDevs();
	if (num_nDevId==0) {
		SendDlgItemMessage(hDlg,hlist,LB_INSERTSTRING,0,
			(LONG)(LPSTR)"gp\MIDIfoCX܂B");
		return;
	}
	for (i=0;i<num_nDevId;i++) {
		midiOutGetDevCaps(i,&moc,sizeof(moc));
		SendDlgItemMessage(hDlg,hlist,LB_INSERTSTRING,i,
			(LONG)(LPSTR)moc.szPname);
	}
}

/**
* @fn GM_systemON_Exclusive_Message
* @brief MIDIfoCXɑ΂GM_systemON_ExclusivebZ[W𑗏o
* @param ()  Ȃ
* @return Ȃ
* @sa 
* @detail
*/
void GM_systemON_Exclusive_Message()
{
	BYTE ExclData[6] = { 0xf0,0x7e,0x7f,0x09,0x01,0xf7 };
	BYTE ExclData2[9] = { 0xf0,0x43,0x10,0x4c,0x00,0x00,0x7e,0x00,0xf7 };

	nDevId=MIDI_device;
	rc=midiOutOpen(&hmo,nDevId,(DWORD)NULL,(DWORD)NULL,CALLBACK_NULL);
	if (rc!=MMSYSERR_NOERROR) return;

    pmh.lpData = (char *)malloc(6);
    pmh.dwBufferLength=6;
    memcpy(pmh.lpData, ExclData, 6);
    rc = midiOutPrepareHeader(hmo, &pmh, sizeof(MIDIHDR));
    if (rc == MMSYSERR_NOERROR)
          rc = midiOutLongMsg(hmo, &pmh, sizeof(MIDIHDR));
	free(pmh.lpData);

    pmh.lpData = (char *)malloc(9);
    pmh.dwBufferLength=9;
    memcpy(pmh.lpData, ExclData2, 9);
    rc = midiOutPrepareHeader(hmo, &pmh, sizeof(MIDIHDR));
    if (rc == MMSYSERR_NOERROR)
          rc = midiOutLongMsg(hmo, &pmh, sizeof(MIDIHDR));
	free(pmh.lpData);
	midiOutClose(hmo);
}

/**
* @fn Change_Program_main
* @brief MIDIfoCXɑ΂ăvOԍAoNԍAʁAo[uAR[Xݒ𑗏o
* @param (chan) ݒ肷`l
* @return Ȃ
* @sa 
* @detail
*/
void  Change_Program_main(DWORD chan)
{
	DWORD com;
	com=0xb0+chan+MIDI_bank1*65536;
	midiOutShortMsg(hmo,com);
	com=0x20+MIDI_bank2*256;
	midiOutShortMsg(hmo,com);
	com=0xC0+chan+(MIDI_program-1)*256;
	midiOutShortMsg(hmo,com);
	com=0xb0+chan+7*256+MIDI_volume*65536;
	midiOutShortMsg(hmo,com);
	com=0xb0+chan+91*256+MIDI_reverb*65536;
	midiOutShortMsg(hmo,com);
	com=0xb0+chan+93*256+MIDI_chorus*65536;
	midiOutShortMsg(hmo,com);
}

/**
* @fn Reset_Pitchbend
* @brief MIDIfoCXɑ΂ăsb`xhEGNXvbVZbg
* @param (chan) ݒ肷`l
* @return Ȃ
* @sa 
* @detail
*/
void  Reset_Pitchbend(DWORD chan)
{
	DWORD com;
// pitch bend reset
	com=0xe0+chan+0x40*65536;
	midiOutShortMsg(hmo,com);

// Add Expression Reset
	com=0xb0+chan+11*256+127*65536;
	midiOutShortMsg(hmo,com);
}

/**
* @fn Change_Program
* @brief MIDIfoCXɑ΂ăvOԍ̐ݒAsb`xhEGNXvbṼZbgsiCj
* @param (chan) ݒ肷`l
* @return Ȃ
* @sa Change_Program_main,Reset_Pitchbend
* @detail
*/
void  Change_Program()
{
	int i;
//	if (max_channel>2) {
	if (max_channel>1) {
		for (i=0;i<max_channel;i++) Change_Program_main(i);
		for (i=0;i<max_channel;i++) Reset_Pitchbend(i);
	}
	else {
		Change_Program_main(0);
		Reset_Pitchbend(0);
	}
}

/**
* @fn Change_Tone
* @brief MIDICxg̃m[gio[wMIDI_tone␳iڒtj
* @param (mess) MIDICxg
* @return ␳MIDICxgf[^i32bitsj
* @sa 
* @detail
*/
DWORD Change_Tone(MIDIMSG mess)
{
	int com,note,note2,vel;
	note2=note=mess.note;
	com=mess.com;
	vel=mess.vel;
	if (com>=0x90 && com<=0x9f) {
		note2=note+MIDI_tone;
		if (note2>127) note2=127;
		if (note2<0) note2=0;
	}
	else if (com>=0x80 && com<=0x8f) {
		note2=note+MIDI_tone;
		if (note2>127) note2=127;
		if (note2<0) note2=0;
	}
	return((DWORD)(com+note2*256+vel*65536));
}

/**
* @fn CloseDevice()
* @brief MIDIfoCXɂMIDIf[^̉tI
* @param () Ȃ
* @return Ȃ
* @sa Close_Draw
* @detail
*/
VOID CloseDevice()
{
	DWORD com;
	int chan;
	if (eStatus != StatusOn) return;
	Sleep(200);
	timeKillEvent(nTimerId);
	timeEndPeriod(nTimerRes);

	for (chan=0;chan<16;chan++) {
		com=0xb0+chan+120*256;
		midiOutShortMsg(hmo,com);
	}
	midiOutReset(hmo);
	midiOutClose(hmo);

	eStatus=StatusOff;

	Close_Draw();
	ReleaseDC(hWnd_main2,hdc_main);

}

/**
* @fn Conv_MIDI_Message
* @brief MIDICxg32bitsMIDIf[^`ɕϊ
* @param (mess) MIDICxg
* @return 32bitsMIDIf[^
* @sa Close_Draw
* @detail
*/
DWORD Conv_MIDI_Message(MIDIMSG mess)
{
	int com,note,vel;
	note=mess.note;
	com=mess.com;
	vel=mess.vel;
	return((DWORD)(com+note*256+vel*65536));
}

/**
* @fn timeproc
* @brief MIDIfoCXɂMIDIf[^̉t𐧌䂷R[obN֐
* @param (nTimerId) VXe^C}[ID
* @param (msg) VXebZ[W
* @param (dwUser) [U
* @param (dwParam1) VXebZ[W̃p[^P
* @param (dwParam2) VXebZ[W̃p[^Q
* @return Ȃ
* @sa Close_Device,Change_Tone,Conv_MIDI_Message
* @detail
*/
VOID CALLBACK timeproc(UINT nTimerId, UINT msg, DWORD dwUser,
		DWORD dwParam1, DWORD dwParam2)
{
	DWORD i,t,td,mess;
	int ch;
	if (MidiSong.dwNextMsg>=song_len)
	{
		CloseDevice();
		MidiSong.dwNextMsg=0;
		MidiSong.ms=0;
		return;
	}
	i=MidiSong.dwNextMsg;
	t=MidiSong.ms*100/MIDI_tempo;
	td=MidiMsg[i].ms*1000/time_unit+500;
	abs_time=MidiMsg[song_len-1].ms;
	total_time=MidiMsg[song_len-1].ms*1000/time_unit;

	while(i<song_len && t>=td)
	{
		if (MIDI_tone!=0) mess=Change_Tone(MidiMsg[i]);
		else mess=Conv_MIDI_Message(MidiMsg[i]);
		ch=mess & 0xf;
		midiOutShortMsg(hmo,mess);

		Draw_Note(hdc_main,td-500,total_time,MidiMsg[i],TRUE);

		MidiSong.dwNextMsg=i+1;
		i++;
		if (i>=song_len) break;
		td=MidiMsg[i].ms*1000/time_unit+500;
	}
	MidiSong.ms +=nTimerRes;
	if (MidiSong.dwNextMsg>=song_len)
	{
		CloseDevice();
		MidiSong.dwNextMsg=0;
		MidiSong.ms=0;
	}
}

/**
* @fn Open_MIDIout
* @brief MIDIfoCXɂMIDIf[^̉tN
* @param () Ȃ
* @return Ȃ
* @sa timeproc,Open_Draw,Draw_Note_bck,Draw_Mess,Change_Program
* @detail
*/
void Open_MIDIout()
{
	DWORD i;
	if (song_len==0) return;

	time_start_disp=0;
	time_end_disp= -1;

	hWnd_main2=GetActiveWindow();
	hdc_main=GetDC(hWnd_main2);
		Open_Draw(hWnd_main2,hdc_main);
		abs_time=MidiMsg[song_len-1].ms;
		for (i=0;i<song_len;i++)
		{
			Draw_Note_bck(hdc_main,MidiMsg[i].ms,abs_time,MidiMsg[i]);
		}
		Draw_Mess();

	MidiSong.dwNextMsg=0;
	MidiSong.ms=0;

	nDevId=MIDI_device;
	rc=midiOutOpen(&hmo,nDevId,(DWORD)NULL,(DWORD)NULL,CALLBACK_NULL);
	if (rc==MMSYSERR_NOERROR) {
		eStatus=StatusOn;
		Change_Program();
	}

	rc=timeGetDevCaps(&tc,sizeof(TIMECAPS));
	if (rc==TIMERR_NOERROR)
	{
		nTimerRes=10;
		timeBeginPeriod(nTimerRes);
//		nTimerId=timeSetEvent(nTimerRes,nTimerRes,timeproc,
//			(DWORD)NULL,TIME_PERIODIC);
//		nTimerId = timeSetEvent(nTimerRes, nTimerRes, (LPTIMECALLBACK)timeproc,
//			(DWORD_PTR)NULL, TIME_PERIODIC);
		nTimerId = timeSetEvent(nTimerRes, nTimerRes, (LPTIMECALLBACK)timeproc,
			(DWORD)NULL, TIME_PERIODIC);
	}
}

/**
* @fn Inquire_Device
* @brief MIDIfoCX̉t̎sԂmF
* @param () Ȃ
* @return s:TRUE, I:FALSE
* @sa
* @detail
*/
BOOL Inquire_Device()
{
	if (eStatus==StatusOn)  return(TRUE);
	else return(FALSE);
}
