/**
* @file MIDcnv_s1_fft.cpp
* @brief MIDI@TuP@g
* @author T.Modegi
* @date  2024.4.24
*/
#include "StdAfx.h"
#include <math.h>
#include <stdio.h>
#include "common.h"
#include "MIDcnv_s1_fft.h"
#include "FFTbase.h"
#include "MIDfile.h"
#include "commparam.h"
#include "commdata.h"

SIGMIDC *sigbuf_c;

SIGMIDC *sigbuf_pre;
int npeak_pre;
int pos_pre;
int fft_step_c;
int diff_level=0;
int window_pos=0;
int num_anal_pos;

/**
* @fn correct_overtone
* @brief g͂ꂽXyNgzɑ΂{␳s
* @param (num) XyNgz̃TCY
* @param (ev) XyNgz
* @return Ȃ
* @sa 
* @detail
*/
/*
void correct_overtone(int num,int ev[])
{
	int i,k,ii;
	double dv,dv2,dvo;
	static int ot[9]={12, 19, 24, 28, 31, 34, 36, 38, 40};
	for (i=0;i<num;i++) {
		dv=dvo=(double)ev[i];
		for (k=0;k<9;k++) {
		ii=i-ot[k];
		if (ii>=0) {
			dv2=(double)ev[ii];
//			dv-=sqrt(dvo*dv2)*(double)overtone_level/100.0;
			dv-=sqrt(dvo*dv2)*(double)overtone_level/50.0/(double)(k+2);
		}
		}
		if (dv<0.0) dv=0.0;
		ev[i]=(int)dv;
	}
}
*/
void correct_overtone(int num,int ev[])
{
	int i,k,ii;
	double dv,dv2;
	static int ot[9]={12, 19, 24, 28, 31, 34, 36, 38, 40};
	for (i=num-13;i>=0;i--) {
		dv=(double)ev[i];
		for (k=0;k<9;k++) {
			ii=i+ot[k];
			if (ii>=num) break;
			dv2=(double)ev[ii];
//			dv-=sqrt(dvo*dv2)*(double)overtone_level/100.0;
			dv2-=sqrt(dv*dv2)*(double)overtone_level/50.0/(double)(k+2);
			if (dv2<0.0) dv2=0.0;
			ev[ii]=(int)dv2;
		}
	}
}

/**
* @fn correct_fundamental
* @brief g͂ꂽXyNgzɑ΂s
* @param (num) XyNgz̃TCY
* @param (ev) ʉa͌̃XyNgz
* @param (evo) ʉa͑ǑXyNgz
* @return Ȃ
* @sa 
* @detail
*/
/*
void correct_fundamental(int num,int ev[],int evo[])
{
	int i,j,jm,vm,k;
	int kk,v;
//	static int ot[9]={12, 19, 24, 28, 31, 34, 36, 38, 40};
	static int ot[30]={
		 0, 12,19,24,28,
		12,  7,12,16,19, 
		19,  5, 9,12,15,
		24,  4, 7,10,12,
		28,  3, 6, 8,10,
		31,  3, 5, 7, 9
	};

	for (i=12;i<num;i++) {
		if (dft_note_tbl[i]>freq_resol_base) break;
		jm=0; vm=0;
		for (j=0;j<max_overtone;j++) {
			v=0;
			for (k=1;k<=4;k++) {
				kk=ot[j*5+k];
				v+=evo[i+kk];
			}
			if (v>vm) {jm=j; vm=v;} 
		}
		j=ot[jm*5];
		if (j>0 && i-j>=0) ev[i-j]+=ev[i];
	}
}
*/
void correct_fundamental(int num,int ev[],int evo[])
{
	int i,j,jm,vm,k;
	int kk,v;
//	static int ot[9]={12, 19, 24, 28, 31, 34, 36, 38, 40, 42, 43};
	static int ot[40]={
		 0, 12,19,24,28,
		12,  7,12,16,19, 
		19,  5, 9,12,15,
		24,  4, 7,10,12,
		28,  3, 6, 8,10,
		31,  3, 5, 7, 9,
		34,  2, 4, 6, 8,
		36,  2, 4, 5, 7 
	};

	for (i=12;i<num;i++) {
		if (dft_note_tbl[i]>freq_resol_base) break;
		jm=0; vm=0;
		for (j=0;j<max_overtone;j++) {
			v=0;
			for (k=1;k<=4;k++) {
				kk=ot[j*5+k];
				v+=evo[i+kk];
			}
			if (v>vm) {jm=j; vm=v;} 
		}
		j=ot[jm*5];
		if (j>0 && i-j>=0) ev[i-j]+=ev[i];
	}
}

/**
* @fn gha_findfreq
* @brief ʉa͂ɂ艹Mt[ɑ΂g͂s
* @param (a) ͑Ώۂ̂Pt[̉Mz
* @param (data)	g͌̉ff[^z
* @param (offset)	͑Ώۂ̐MfЂ̃Tvʒu
* @param (dt1)	Ỏ̓t[Ƃ̃I[o[bvԁi炩ߑւvZς݂̃Tvj
* @param (dt2) 㑱̉̓t[ƃI[o[bvȂŐԁiTvj
* @param (mode) 0:Ỏ̓t[̌ʂƕ]AωΉ͂𒆎~, 1:ɉ͎s
* @param (prev)@Ỏ̓t[ɂg͌̃XyNgz
* @return ͂ꂽff[^
* @sa calc_correlationA,calc_correlation,calc_correlationD,correct_overtone,correct_fundamental
* @detail
*/
int gha_findfreq(short a[],SIGMIDC data[],int offset,int dt1,int dt2,int mode,int prev[])
{
	int i,j;
	int jm;
	int v,sum;
	int ev[128];
	int evo[128];
	int sv,cv;
	int sub[128];
	int subo[128];
	BOOL sel[128];
	int conn[128];
	int maxval;
	double dev;
	int iev;
	int n0,n1,n2,nn,dn0,dn1,dn2,dl0,dl1,dl2,adl0,adl1,adl2;

//	Descrete Fourier Transform
	for (i=0;i<fft_size;i++) ift_work[i]=(int)a[i];
//	ampl_weight_func();

	sum=0;
	for (i=0;i<dft_note_cnt;i++) {
		sub[i]=freq_resol[i]/2;
		evo[i]=0;
		for (j=0;j<freq_resol[i];j++) {
			iev=calc_correlationA(i,j,offset,dt1,dt2,&sv,&cv);
			if (iev>evo[i]) {sub[i]=j; evo[i]=iev;}
		}
		v=evo[i]-prev[i];
//		if (v<0) v=0;
		if (v<0) v= -v;
		if (evo[i]+prev[i]>0) v=v*1000/(evo[i]+prev[i]);
		else if (v>0) v=500;
		sum+=v;
		prev[i]=evo[i];
	}
	sum=sum/dft_note_cnt;


// Micro-tone 
	for (i=0;i<dft_note_cnt;i++) {
		subo[i]=sub[i];
		if (freq_resol[i]>1 && freq_resol[i]<freq_resol_def) {
			subo[i] -=freq_resol[i]/2;
			subo[i]=subo[i]*freq_resol_def/freq_resol[i];
			subo[i] +=freq_resol_def/2;
			if (subo[i]>=freq_resol_def) subo[i]=freq_resol_def-1;
			if (subo[i]<0) subo[i]=0;
		}
	}

// Determine whether proceeding to GHA Processing
	for (i=0;i<dft_note_cnt;i++) conn[i]=0;
	if (mode==2 || (mode==0 && sum<variable_rate)) return(-1);

//
//  GHA Process
//
	for (j=0;j<dft_note_cnt;j++) {ev[j]=sel[j]=0;}
	for (i=0;i<dft_note_cnt;i++) {
		maxval=0;
		jm= -1;
		for (j=0;j<dft_note_cnt;j++) {
			if (sel[j]==0) {
				if (evo[j]>maxval) {
					maxval=evo[j];
					jm=j;
				}
			}
		}
		if (jm== -1)  break;
		iev=calc_correlation(jm,sub[jm],&sv,&cv);
		ev[jm] +=iev;
		calc_correlationD(jm,sub[jm],sv,cv);
		sel[jm]=1;
	}

// calculate power
	if (fundamental_correct) correct_fundamental(dft_note_cnt,ev,evo);
	if (overtone_correct) correct_overtone(dft_note_cnt,ev);

	for (i=0;i<dft_note_cnt;i++) {
		dev=(double)ev[i]/16384.0/16384.0;
		ev[i]=(int)(pow(dev,0.25)*16384.0);
	}

// Determine Connectivity
	for (i=0;i<dft_note_cnt;i++) {
		conn[i]=0;
			dl0=ev[i]-prev[i+256];
			if (dl0<0) dl0= -dl0;
			if (ev[i]+prev[i+256]>0) dl0=dl0*100/(ev[i]+prev[i+256]);
			else if (dl0>0) dl0=50;
			adl0=dl0;
			nn=dft_note_tbl[i]*freq_resol_def+subo[i];
			n0=dft_note_tbl[i]*freq_resol_def+prev[i+dft_note_cnt];
			dn0=n0-nn;
			if (dn0<0)  dn0= -dn0;
			if (adl0<slice_difflevel && dn0<=slice_diffnote) conn[i]=1;

			n1=n2= -1;
			dl1=dl2=99999;
			if (i>0) {
				n1=dft_note_tbl[i-1]*freq_resol_def+prev[i-1+dft_note_cnt];
				dn1=n1-nn;
				if (dn1<0)  dn1= -dn1;
				dl1=ev[i]-prev[i-1+256];
				if (dl1<0) dl1= -dl1;
				if (ev[i]+prev[i-1+256]>0) dl1=dl1*100/(ev[i]+prev[i-1+256]);
				else if (dl1>0) dl1=50;
				adl1=dl1;
				if (adl1<slice_difflevel && dn1<=slice_diffnote) {
					if (conn[i]==0) conn[i]=2;
					else if (conn[i]==1 && adl1<adl0) conn[i]=2;
				}
			}
			if (i<dft_note_cnt-1) {
				n2=dft_note_tbl[i+1]*freq_resol_def+prev[i+1+dft_note_cnt];
				dn2=n2-nn;
				dl2=ev[i]-prev[i+1+256];
				if (dl2<0) dl2= -dl2;
				if (ev[i]+prev[i+1+256]>0) dl2=dl2*100/(ev[i]+prev[i+1+256]);
				else if (dl2>0) dl2=50;
				adl2=dl2;
				if (adl2<slice_difflevel && dn2<=slice_diffnote) {
					if (conn[i]==0) conn[i]=3;
					else if (conn[i]==1 && adl2<adl0) conn[i]=3;
					else if (conn[i]==2 && adl2<adl1) conn[i]=3;
				}
			}
		prev[i+256]=ev[i];
		prev[i+dft_note_cnt]=subo[i];
	}


// Select Candidates
//
//  Selection
//
	for (i=0;i<dft_note_cnt;i++) {
		data[i].note=dft_note_tbl[i];
		data[i].level=ev[i];
		data[i].sub=subo[i];
		data[i].conn=conn[i];
	}

	return (dft_note_cnt);
}

/**
* @fn gha_findfreq2
* @brief ʉa͂ɂ艹Mt[ɑ΂g͂sidt[́j
* @param (datap) Ỏ̓t[ɂg͌̂̉ff[^z
* @param (data)	g͌̂̉ff[^z
* @param (dt)	Ỏ̓t[Ƃ̎ԍiTvj
* @return ͂ꂽff[^
* @sa 
* @detail
*/
int gha_findfreq2(SIGMIDC datap[],SIGMIDC data[],int dt)
{
	int i;
	int ev[128];
	double dev,devp;
	int tcycle,ifa;

// calculate power
	for (i=0;i<dft_note_cnt;i++) {
		ifa=freq_adr[i]+datap[i].sub;
		if (cycle_tbl[ifa]<0.0) tcycle=0;
		else  tcycle=width_tbl[ifa];
		ev[i]=0;
		if (tcycle>0) {
			devp=pow((double)datap[i].level,4.0);
			dev=pow((double)data[i].level,4.0);
			devp=(devp*(double)tcycle-sqrt(devp*dev)*(double)(tcycle-dt))/(double)dt;
			if (devp>0.0) ev[i]=(int)pow(devp,0.25);
		}
	}

// Select Candidates
//
//  Selection
//
	for (i=0;i<dft_note_cnt;i++) {
		datap[i].level=ev[i];
	}

	return (dft_note_cnt);
}

/**
* @fn save_note_interval
* @brief note_interval[]zFFTt@Cɕۑ
* @param (*fp) ۑFFTeLXg`t@C̃t@C|C^ 
* @return Ȃ
* @sa 
* @detail
*/
void save_note_interval(FILE *fp)
{
	int i,v;
	if (variable_interval) v=1; else v=0;
	fprintf(fp,"%d\n",v);
	for (i=0;i<16;i++) {
		fprintf(fp,"%d %d %d %d %d %d %d %d\n",
		    note_interval[i*8],note_interval[i*8+1],note_interval[i*8+2],note_interval[i*8+3],
		    note_interval[i*8+4],note_interval[i*8+5],note_interval[i*8+6],note_interval[i*8+7]);
	}
}

/**
* @fn fMIDcnv_s1_fft
* @brief ʉa͂ɂ艹MŜɑ΂g͂siCj
* @param () Ȃ
* @return ͂ꂽff[^
* @sa gha_findfreq,gha_findfreq2,disp_progress
* @detail@͂ꂽẢff[^FFT`g̓f[^̃eLXgt@C
iWAVt@C.fftjɕۑ
*/
int fMIDcnv_s1_fft()
{ 
	int i,p,j;
	int cnt,npeak;
	int level;
	int dt1,dt2;
	int *prev;
	int mode;
	FILE *fft_fd;
	unsigned int per;
	int dt;

	max_fftlevel=0;
	Change_FileExt(filename,filename2,"fft");
	fft_fd=fopen(filename2,"w");
	if (fine_note) p=1; else p=0;
	fprintf(fft_fd,"%d %d %d\n",p,slice_difflevel,slice_diffnote);
	
	set_freq_resol();
	dft_table();
	sincos_tbl();
	save_note_interval(fft_fd);

	sigbuf_c=(SIGMIDC *)malloc(dft_note_cnt*sizeof(SIGMIDC));
	sigbuf_pre=(SIGMIDC *)malloc(dft_note_cnt*sizeof(SIGMIDC));
	npeak_pre=0;

	cnt=0;
	p=pos_pre=0;
	num_anal_pos=(num_sample-fft_size)/fft_step;

	prev=(int *)malloc(768*4);
	for (i=0;i<768;i++) prev[i]=0;

	i=0;
	while (i<num_anal_pos) {
		window_pos=p;
		if (p>=num_sample-fft_size) break;

		mode=0;
		if (i==0)  {dt1=0; mode=1;}
		else  dt1=fft_step;
		if (i==num_anal_pos-1)  {dt2=0; mode=1;}
		else  dt2=fft_step;

		if (variable_step==FALSE || p-pos_pre>=fft_size/2) mode=1;
		npeak=gha_findfreq(&wavbuf[p],sigbuf_c,p,dt1,dt2,mode,prev);

		if (npeak>0) {
			if (p>pos_pre) {
				dt=(p-pos_pre)*note_length_ratio/100;
				if (double_frame_anal==TRUE && p>=fft_size) {
				    gha_findfreq2(sigbuf_pre,sigbuf_c,dt);
				}
				fprintf(fft_fd,"%d %ld %d\n",npeak_pre,pos_pre,dt);
				for (j=0;j<npeak_pre;j++) {
					level=sigbuf_pre[j].level;
					if (level>max_fftlevel)  max_fftlevel=level;
					fprintf(fft_fd,"%d %d %d %d\n",
						sigbuf_pre[j].note,sigbuf_pre[j].sub,level,sigbuf_pre[j].conn);
				}
				cnt +=npeak_pre;
			}
			pos_pre=p;
			npeak_pre=npeak;
			for (j=0;j<npeak;j++)  {
				sigbuf_pre[j].level=sigbuf_c[j].level;
				sigbuf_pre[j].note=sigbuf_c[j].note;
				sigbuf_pre[j].sub=sigbuf_c[j].sub;
				sigbuf_pre[j].conn=sigbuf_c[j].conn;
			}
		}
		if (i%10==0)  {
			per=(unsigned int)((double)p*100.0/(double)num_sample);
			disp_progress(2,per);
		}
			i++;
			p +=fft_step;
		if (mess_file_close) break;
	}
	dft_close();
	fclose(fft_fd);
	free(sigbuf_c);
	free(sigbuf_pre);
	free(prev);

	num_record=cnt;

	return(num_record);
}
