/**
* @file MIDcnv_s4_mid.cpp
* @brief MIDI@TuS@MIDICxg
* @author T.Modegi
* @date  2020.1.22
*/
#include "StdAfx.h"
#include "common.h"
#include "commparam.h"
#include "MIDfile.h"
#include "MIDdraw.h"
#include "MIDcnv_s1_fft.h"
#include "MIDcnv_s4_mid.h"
#include "commdata.h"

#define TIME_DIV 1000

/**
* @fn MIDImess_decode
* @brief 32bitsMIDICxgf[^eCxgp[^𒊏o
* @param (data) 32bitsMIDI`Cxgf[^
* @param (note) MIDIm[gio[
* @param (vel) MIDIxVeB
* @param (sw) MIDICxgʃR}h
* @return Ȃ
* @sa 
* @detail
*/
void MIDImess_decode(DWORD data,int *note,int *vel,int *sw)
{
	int dat,v,n;
	dat=(int)data;
	if (dat==0) { *note= *vel= *sw= -1; return; }
	v=dat/65536;
	n=(dat-v*65536)/256;
	*sw=dat-v*65536-n*256;
	*note=n;
	*vel=v;
}

/**
* @fn MIDImess_encode
* @brief MIDICxgp[^32bitsMIDI`Cxgf[^ɕ
* @param (note) MIDIm[gio[
* @param (vel) MIDIxVeB
* @param (sw) MIDICxgʃR}h
* @return ꂽ32bitsMIDI`Cxgf[^
* @sa 
* @detail
*/
DWORD MIDImess_encode(int note,int vel,int sw)
{
	return((DWORD)(sw+note*256+vel*65536));
}

/**
* @fn count_harmony
* @brief AMIDICxgf[^őaZo
* @param () Ȃ
* @return őa
* @sa 
* @detail
*/
int count_harmony()
{
	int i;
	int c;
	int t,pt;
	int sw;
	int cmax;

	cmax=0;
	t=pt=MidiMsg[0].ms;
	c=0;
	for (i=0;i<(int)song_len;i++) {
		t=MidiMsg[i].ms;
		if (t>pt) if (c>cmax) cmax=c;
		sw=MidiMsg[i].com;
		if (sw>=0x90 && sw<=0x9f) c++;
		else if (sw>=0x80 && sw<=0x8f) c--;
		pt=t;
	}
	return(cmax);
}

/**
* @fn count_bitrate
* @brief AMIDICxgf[^őrbg[gZo
* @param () Ȃ
* @return őrbg[g
* @sa 
* @detail
*/
int count_bitrate()
{
	DWORD i,is,ie;
	DWORD tmax,tm,tdis,n,num;
	int br,mbr;

	tm=MidiMsg[song_len-1].ms;
	num=tm/TIME_DIV;
	if (num==0) num=1;
	tdis=TIME_DIV;

	mbr=0;
	is=0;
	for (n=0;n<num;n++) {
		if (n==num-1) tmax=tm;
		else tmax=(n+1)*TIME_DIV;
		i=is;
		ie=is;
		while (i<song_len) {
			if (MidiMsg[i].ms>tmax || i==song_len-1) {
				if (i>is)  ie=i-1;
				else ie=is;
				if (i==song_len-1) {
					ie=i;
					tdis=tmax-n*TIME_DIV;
				}
				if (ie>is) {
					br=(int)((double)(ie-is+1)*40000.0/(double)tdis);
					if (br>mbr)  mbr=br;
					is=ie+1;
				}
				break;
			}
			i++;
		}
	}
	return(mbr);
}

/**
* @fn remove_MIDIevent
* @brief MIDICxgf[^폜AMIDICxgzXV
* @param () Ȃ
* @return XVꂽMIDICxg
* @sa 
* @detail@MIDICxgʃR}hOɂȂĂf[^폜
*/
int remove_MIDIevent()
{
	int i,irec;
	irec=0;
	for (i=0;i<(int)song_len;i++) {
		if (MidiMsg[i].com!=0) {
			MidiMsg[irec].ms=MidiMsg[i].ms;
			MidiMsg[irec].com=MidiMsg[i].com;
			MidiMsg[irec].note=MidiMsg[i].note;
			MidiMsg[irec].vel=MidiMsg[i].vel;
			MidiMsg[irec].chan=MidiMsg[i].chan;
			irec++;
		}
	}
	return (irec);
}

/**
* @fn adjust_harmony
* @brief őaȉɂȂ悤ɒ
* @param () Ȃ
* @return XVꂽMIDICxg
* @sa emove_MIDIevent
* @detail@őamax_harmonyɐ
*/
int adjust_harmony()
{
	int irec,note,vel,sw;
	int irec2,note2,vel2,sw2;
	int n,i,ir,ir2;
	int c1,c2;
	int onrec[128];
	int onvel[128];
	int ondt[128];
	int offrec[128];
	int mh;
	int v,vmin,im;

	for (n=0;n<128;n++) { onrec[n]=onvel[n]=ondt[n]=offrec[n]= -1; }
	mh=0;
	irec=0;
	while(irec<(int)song_len) {
		sw=MidiMsg[irec].com; note=MidiMsg[irec].note; vel=MidiMsg[irec].vel;
		if (sw>=0x90 && sw<=0x9f) {
			onrec[note]=irec;
			onvel[note]=vel;
			irec2=irec+1;
			while(irec2<(int)song_len) {
				sw2=MidiMsg[irec2].com; note2=MidiMsg[irec2].note; vel2=MidiMsg[irec2].vel;
				if (sw2>=0x80 && sw2<=0x8f && note2==note) {
					ondt[note]=MidiMsg[irec2].ms-MidiMsg[irec].ms;
					offrec[note]=irec2;
					mh++;
					break;
				}
				irec2++;
			}
			
			if (mh>max_harmony) {
				im= -1;
				vmin= -1;
				for (i=0;i<128;i++) {
					if (onrec[i]>=0) {
						v=onvel[i]*ondt[i];
						if (vmin== -1 || v<vmin) {
							im=i;
							vmin=v;
						}
					}
				}
				ir=onrec[im];
				ir2=offrec[im];

				if (max_channel>2) {
					sw=MidiMsg[ir].com;
					c1=sw & 0x0f;
					i=ir+1;
					while (i<ir2) {
					    sw=MidiMsg[i].com;
						c2=sw & 0x0f;
						if ((sw<0x80 || sw>0x9f) && c2==c1) MidiMsg[i].com=0;
						i++;
					}
				}
					MidiMsg[ir].com=0;
					MidiMsg[ir2].com=0;

				onrec[im]=onvel[im]=ondt[im]=offrec[im]= -1;
				mh--;
			}
		}
		else if (sw>=0x80 && sw<=0x8f) {
			if (onrec[note]>=0) {
			onrec[note]=onvel[note]=ondt[note]=offrec[note]= -1;
			mh--;
			}
		}
		irec++;
	}

// Remove Notes
	song_len=remove_MIDIevent();

	return(song_len);
}

/**
* @fn adjust_bitrate
* @brief őrbg[gȉɂȂ悤ɒ
* @param () Ȃ
* @return XVꂽMIDICxg
* @sa emove_MIDIevent
* @detail@őrbg[gmax_bitrateɐ
*/
int adjust_bitrate()
{
	int i,j;
	int irec,irec2;
	int note,vel,sw,sw2;
	int c1,c2;
	int dt,n1,v1;
	NOTE_PRIORITY *priorbuf,tmp;
	int num_priorbuf,num;

	int is,ie;
	int tmax,tm,tdis,n,nm;
	int cnt;

	priorbuf=(NOTE_PRIORITY *)malloc(song_len*sizeof(NOTE_PRIORITY)/2);

	tm=MidiMsg[(int)song_len-1].ms;
	nm=tm/TIME_DIV;
	if (nm==0) nm=1;
	tdis=TIME_DIV;

	is=0;
	for (n=0;n<nm;n++) {
		if (n==nm-1) {
			tmax=tm;
			tdis=tmax-n*TIME_DIV;
		}
		else tmax=(n+1)*TIME_DIV;
		ie=is;
		i=is;
		while (i<(int)song_len) {
			if ((int)MidiMsg[i].ms>tmax || i==(int)song_len-1) {
				if (i>is) ie=i-1;
				else ie=is;
				if (i==(int)song_len-1) ie=i;
				break;
			}
			i++;
		}
		if (ie>is) {

// LOOP START
// Set Array
	irec=is;
	num_priorbuf=0;
	cnt=0;
	while(irec<=ie) {
		sw=MidiMsg[irec].com; note=MidiMsg[irec].note; vel=MidiMsg[irec].vel;
		if (MidiMsg[irec].com!=0)  cnt++;
		if (sw>=0x90 && sw<=0x9f) {
			n1=note; v1=vel; c1=sw & 0x0f;
			irec2=irec+1;
			dt= -1;
			while(irec2<(int)song_len) {
			    sw2=MidiMsg[irec2].com; note=MidiMsg[irec2].note; vel=MidiMsg[irec2].vel;
				c2=sw2 & 0x0f;
				if (sw2>=0x80 && sw2<=0x8f && c1==c2 && note==n1) {
					dt=MidiMsg[irec2].ms-MidiMsg[irec].ms;
					break;
				}
				irec2++;
			}
			if (dt>0) {
				priorbuf[num_priorbuf].pos=irec;
				priorbuf[num_priorbuf].offpos=irec2;
				priorbuf[num_priorbuf].value=v1*dt;
				num_priorbuf++;
			}
		}
		irec++;
	}

// Sort by value
	for (i=0;i<num_priorbuf;i++) {
		for (j=i+1;j<num_priorbuf;j++) {
			if (priorbuf[j].value>priorbuf[i].value) {
				tmp.pos=priorbuf[i].pos;
				tmp.offpos=priorbuf[i].offpos;
				tmp.value=priorbuf[i].value;
				priorbuf[i].pos=priorbuf[j].pos;
				priorbuf[i].offpos=priorbuf[j].offpos;
				priorbuf[i].value=priorbuf[j].value;
				priorbuf[j].pos=tmp.pos;
				priorbuf[j].offpos=tmp.offpos;
				priorbuf[j].value=tmp.value;
			}
		}
	}

// Remove Notes 1
	num=(int)((double)max_bitrate*(double)tdis/40000.0);
	for (i=num_priorbuf-1;i>=0;i--) {
		if (cnt<=num) break;
		irec=priorbuf[i].pos;
		irec2=priorbuf[i].offpos;
		if (max_channel>2) {
		    sw=MidiMsg[irec].com;
			c1=sw & 0x0f;
			j=irec+1;
			while (j<irec2) {
		        sw=MidiMsg[j].com;
				c2=sw & 0x0f;
				if ((sw>=0xb0) && c2==c1) {
					MidiMsg[j].com=0;
					if (j<=ie) cnt--;
				}
				j++;
			}
		}
		MidiMsg[irec].com=0;
		MidiMsg[irec2].com=0;
		cnt--;  if (irec2<=ie)  cnt--;
	}

// LOOP END
		is=ie+1;
		}
	}
	free(priorbuf);

// Remove Notes
	song_len=remove_MIDIevent();

	return(song_len);
}
